/**
 * @file H26XDeviceSource.cpp
 * @author 吴晨
 * @brief 适配 live555
 * 
 * @copyright Copyright (c) 2022-2024 OurEDA
 */

#include "H26XDeviceSource.hh"

#include <cstdlib>
#include <cstring>
#include "GroupsockHelper.hh"

unsigned H26XDeviceSource::referenceCount = 0;
EventTriggerId H26XDeviceSource::eventTriggerId = 0;

H26XDeviceSource *H26XDeviceSource::createNew(UsageEnvironment &env, Boolean isH265) {
    return new H26XDeviceSource(env, isH265);
}

H26XDeviceSource::H26XDeviceSource(UsageEnvironment &env, Boolean isH265)
    : FramedSource(env), isH265(isH265), frameDataQueue(3) {
    if (eventTriggerId == 0) {
        eventTriggerId = envir().taskScheduler().createEventTrigger([](void *clientData) {
            ((H26XDeviceSource *) clientData)->deliverFrame();
        });
    }
    ++referenceCount;
}

H26XDeviceSource::~H26XDeviceSource() {
    --referenceCount;
    if (referenceCount == 0) {
        envir().taskScheduler().deleteEventTrigger(eventTriggerId);
        eventTriggerId = 0;
    }
    FrameData frameData;
    while (frameDataQueue.try_dequeue(frameData)) {
        delete[] frameData.data;
    }
}

codec_type_t H26XDeviceSource::getCodecType() {
    return isH265 ? H265 : H264;
}

void H26XDeviceSource::signalNewFrame(const FrameData &frameData) {
    if (!frameDataQueue.try_enqueue(frameData)) {
        envir() << "Stream data queue full!\n";
        delete[] frameData.data;
    }
    envir().taskScheduler().triggerEvent(eventTriggerId, this);
}

unsigned H26XDeviceSource::maxFrameSize() const {
    return 0;
}

void H26XDeviceSource::doGetNextFrame() {
    if (frameDataQueue.size_approx() > 0) {
        deliverFrame();
    }
}

void H26XDeviceSource::deliverFrame() {
    // 是否准备好接收数据
    if (!isCurrentlyAwaitingData()) return;
    // 从队列里取出数据
    FrameData buffer;
    if (!frameDataQueue.try_dequeue(buffer)) return;
    // 处理大小
    if (buffer.length > fMaxSize) {
        fFrameSize = fMaxSize;
        fNumTruncatedBytes = buffer.length - fMaxSize;
    } else {
        fFrameSize = buffer.length;
    }
    // 设置这帧的时间码
    if (buffer.timestamp == 0) {
        gettimeofday(&fPresentationTime, NULL);
    } else {
        fPresentationTime.tv_sec = buffer.timestamp / 1000000;
        fPresentationTime.tv_usec = buffer.timestamp % 1000000;
    }
    // 拷贝码流
    // 去掉start code 0x00000001
    int offset = buffer.data[0] == 0 && buffer.data[1] == 0 && buffer.data[2] == 0 && buffer.data[3] == 1 ? 4 : 0;
    fFrameSize -= offset;
    memcpy(fTo, buffer.data + offset, fFrameSize);
    delete[] buffer.data;
    // 通知live555
    afterGetting(this);
}
