/**
 * @file MJPEGDeviceSource.cpp
 * @author 吴晨
 * @brief 适配 live555
 *
 * @copyright Copyright (c) 2022-2024 OurEDA
 */

#include "MJPEGDeviceSource.hh"

#include <cstdlib>
#include <cstring>
#include "GroupsockHelper.hh"

unsigned MJPEGDeviceSource::referenceCount = 0;
EventTriggerId MJPEGDeviceSource::eventTriggerId = 0;

MJPEGDeviceSource *MJPEGDeviceSource::createNew(UsageEnvironment &env) {
    return new MJPEGDeviceSource(env);
}

MJPEGDeviceSource::MJPEGDeviceSource(UsageEnvironment &env)
    : JPEGVideoSource(env), parser(), frameDataQueue(3) {
    if (eventTriggerId == 0) {
        eventTriggerId = envir().taskScheduler().createEventTrigger([](void *clientData) {
            ((MJPEGDeviceSource *) clientData)->deliverFrame();
        });
    }
    ++referenceCount;
}

MJPEGDeviceSource::~MJPEGDeviceSource() {
    --referenceCount;
    if (referenceCount == 0) {
        envir().taskScheduler().deleteEventTrigger(eventTriggerId);
        eventTriggerId = 0;
    }
    FrameData frameData;
    while (frameDataQueue.try_dequeue(frameData)) {
        delete[] frameData.data;
    }
}

codec_type_t MJPEGDeviceSource::getCodecType() {
    return ROV_JPEG;
}

void MJPEGDeviceSource::signalNewFrame(const FrameData &frameData) {
    if (!frameDataQueue.try_enqueue(frameData)) {
        envir() << "Stream data queue full!\n";
        delete[] frameData.data;
    }
    envir().taskScheduler().triggerEvent(eventTriggerId, this);
}

unsigned MJPEGDeviceSource::maxFrameSize() const {
    return 0;
}

void MJPEGDeviceSource::doGetNextFrame() {
    if (frameDataQueue.size_approx() > 0) {
        deliverFrame();
    }
}

void MJPEGDeviceSource::deliverFrame() {
    if (!isCurrentlyAwaitingData()) return;

    FrameData buffer;
    if (!frameDataQueue.try_dequeue(buffer)) return;

    if (parser.parse(buffer.data, buffer.length) != 0) {
        delete[] buffer.data;
        return;
    }

    if (buffer.length > fMaxSize) {
        fFrameSize = fMaxSize;
        fNumTruncatedBytes = buffer.length - fMaxSize;
    } else {
        fFrameSize = buffer.length;
    }

    if (buffer.timestamp == 0) {
        gettimeofday(&fPresentationTime, NULL);
    } else {
        fPresentationTime.tv_sec = buffer.timestamp / 1000000;
        fPresentationTime.tv_usec = buffer.timestamp % 1000000;
    }

    memcpy(fTo, buffer.data, fFrameSize);
    delete[] buffer.data;

    afterGetting(this);
}
